@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 2004-2014 Open Source Applications Foundation.
+ * Copyright (c) 2004-2018 Open Source Applications Foundation.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -187,6 +187,10 @@ static PyObject *t_messageformat_setFormats(t_messageformat *self,
                                             PyObject *arg);
 static PyObject *t_messageformat_setFormat(t_messageformat *self,
                                            PyObject *args);
+#if U_ICU_VERSION_HEX >= 0x04000000
+static PyObject *t_messageformat_getFormatNames(t_messageformat *self);
+#endif
+
 static PyObject *t_messageformat_format(t_messageformat *self, PyObject *args);
 static PyObject *t_messageformat_parse(t_messageformat *self, PyObject *args);
 static PyObject *t_messageformat_formatMessage(PyTypeObject *type,
@@ -197,8 +201,13 @@ static PyNumberMethods t_messageformat_as_number = {
     0,                                 /* nb_add */
     0,                                 /* nb_subtract */
     0,                                 /* nb_multiply */
+#if PY_MAJOR_VERSION >= 3
+    (binaryfunc) t_messageformat_mod,  /* nb_remainder */
+    0,                                 /* nb_divmod */
+#else
     0,                                 /* nb_divide */
     (binaryfunc) t_messageformat_mod,  /* nb_remainder */
+#endif
 };
 
 static PyMethodDef t_messageformat_methods[] = {
@@ -209,6 +218,9 @@ static PyMethodDef t_messageformat_methods[] = {
     DECLARE_METHOD(t_messageformat, getFormats, METH_NOARGS),
     DECLARE_METHOD(t_messageformat, setFormats, METH_O),
     DECLARE_METHOD(t_messageformat, setFormat, METH_VARARGS),
+#if U_ICU_VERSION_HEX >= 0x04000000
+    DECLARE_METHOD(t_messageformat, getFormatNames, METH_NOARGS),
+#endif
     DECLARE_METHOD(t_messageformat, format, METH_VARARGS),
     DECLARE_METHOD(t_messageformat, parse, METH_VARARGS),
     DECLARE_METHOD(t_messageformat, formatMessage, METH_VARARGS | METH_CLASS),
@@ -347,6 +359,65 @@ DECLARE_TYPE(ListFormatter, t_listformatter, UObject, ListFormatter,
 
 #endif
 
+#if U_ICU_VERSION_HEX >= VERSION_HEX(57, 0, 0)
+
+class t_simpleformatter : public _wrapper {
+public:
+    SimpleFormatter *object;
+    PyObject *pattern;
+};
+
+static int t_simpleformatter_init(t_simpleformatter *self,
+                                  PyObject *args, PyObject *kwds);
+
+static PyObject *t_simpleformatter_applyPattern(
+    t_simpleformatter *self, PyObject *arg);
+static PyObject *t_simpleformatter_applyPatternMinMaxArguments(
+    t_simpleformatter *self, PyObject *args);
+static PyObject *t_simpleformatter_getArgumentLimit(
+    t_simpleformatter *self, PyObject *args);
+static PyObject *t_simpleformatter_format(
+    t_simpleformatter *self, PyObject *args);
+static PyObject *t_simpleformatter_formatStrings(
+    t_simpleformatter *self, PyObject *arg);
+
+static PyNumberMethods t_simpleformatter_as_number = {
+    0,                                             /* nb_add */
+    0,                                             /* nb_subtract */
+    0,                                             /* nb_multiply */
+#if PY_MAJOR_VERSION >= 3
+    (binaryfunc) t_simpleformatter_formatStrings,  /* nb_remainder */
+    0,                                             /* nb_divmod */
+#else
+    0,                                             /* nb_divide */
+    (binaryfunc) t_simpleformatter_formatStrings,  /* nb_remainder */
+#endif
+};
+
+static PyMethodDef t_simpleformatter_methods[] = {
+    DECLARE_METHOD(t_simpleformatter, applyPattern, METH_O),
+    DECLARE_METHOD(t_simpleformatter, applyPatternMinMaxArguments, METH_VARARGS),
+    DECLARE_METHOD(t_simpleformatter, getArgumentLimit, METH_NOARGS),
+    DECLARE_METHOD(t_simpleformatter, format, METH_VARARGS),
+    DECLARE_METHOD(t_simpleformatter, formatStrings, METH_O),
+    { NULL, NULL, 0, NULL }
+};
+
+static void t_simpleformatter_dealloc(t_simpleformatter *self)
+{
+    if (self->flags & T_OWNED)
+        delete self->object;
+    self->object = NULL;
+
+    Py_CLEAR(self->pattern);
+    Py_TYPE(self)->tp_free((PyObject *) self);
+}
+
+DECLARE_TYPE(SimpleFormatter, t_simpleformatter, UMemory, SimpleFormatter,
+             t_simpleformatter_init, t_simpleformatter_dealloc);
+
+#endif
+
 /* FieldPosition */
 
 static int t_fieldposition_init(t_fieldposition *self,
@@ -372,7 +443,7 @@ static int t_fieldposition_init(t_fieldposition *self,
         PyErr_SetArgsError((PyObject *) self, "__init__", args);
         return -1;
     }
-        
+
     if (self->object)
         return 0;
 
@@ -486,7 +557,7 @@ static int t_parseposition_init(t_parseposition *self,
         PyErr_SetArgsError((PyObject *) self, "__init__", args);
         return -1;
     }
-        
+
     if (self->object)
         return 0;
 
@@ -561,8 +632,7 @@ static PyObject *t_parseposition_richcmp(t_parseposition *self, PyObject *arg, i
 
 PyObject *t_format_format(t_format *self, PyObject *args)
 {
-    UnicodeString *u;
-    UnicodeString _u;
+    UnicodeString *u, _u;
     Formattable *obj;
     FieldPosition *fp;
 
@@ -787,7 +857,7 @@ static int t_timeunitformat_init(t_timeunitformat *self,
         PyErr_SetArgsError((PyObject *) self, "__init__", args);
         return -1;
     }
-        
+
     if (self->object)
         return 0;
 
@@ -862,7 +932,7 @@ static int t_messageformat_init(t_messageformat *self,
         PyErr_SetArgsError((PyObject *) self, "__init__", args);
         return -1;
     }
-        
+
     if (self->object)
         return 0;
 
@@ -949,8 +1019,16 @@ static PyObject *t_messageformat_getFormats(t_messageformat *self)
     PyObject *list = PyList_New(count);
 
     for (int i = 0; i < count; i++) {
-        PyObject *obj = wrap_Format(formats[i]->clone());
-        PyList_SET_ITEM(list, i, obj);
+        if (formats[i] == NULL)
+        {
+            PyList_SET_ITEM(list, i, Py_None);
+            Py_INCREF(Py_None);
+        }
+        else
+        {
+            PyObject *obj = wrap_Format(formats[i]->clone());
+            PyList_SET_ITEM(list, i, obj);
+        }
     }
 
     return list;
@@ -1005,7 +1083,7 @@ static PyObject *t_messageformat_format(t_messageformat *self, PyObject *args)
                     self->object->format(f, len, _u, _fp, status);
                     delete[] f;
                 });
-                
+
             return PyUnicode_FromUnicodeString(&_u);
         }
         break;
@@ -1022,7 +1100,7 @@ static PyObject *t_messageformat_format(t_messageformat *self, PyObject *args)
                     self->object->format(f, len, _u, *fp, status);
                     delete[] f;
                 });
-                
+
             return PyUnicode_FromUnicodeString(&_u);
         }
         if (!parseArgs(args, "RU", TYPE_CLASSID(Formattable),
@@ -1077,11 +1155,23 @@ static PyObject *t_messageformat_format(t_messageformat *self, PyObject *args)
     return t_format_format((t_format *) self, args);
 }
 
+#if U_ICU_VERSION_HEX >= 0x04000000
+
+static PyObject *t_messageformat_getFormatNames(t_messageformat *self)
+{
+    StringEnumeration *se;
+    STATUS_CALL(se = self->object->getFormatNames(status));
+
+    return wrap_StringEnumeration(se, T_OWNED);
+}
+
+#endif
+
 static PyObject *fromFormattableArray(Formattable *formattables,
                                       int len, int dispose)
 {
     PyObject *list = PyList_New(len);
-    
+
     for (int i = 0; i < len; i++)
         PyList_SET_ITEM(list, i, wrap_Formattable(formattables[i]));
 
@@ -1119,7 +1209,7 @@ static PyObject *t_messageformat_parse(t_messageformat *self, PyObject *args)
         }
         break;
     }
-    
+
     return PyErr_SetArgsError((PyObject *) self, "parse", args);
 }
 
@@ -1185,7 +1275,7 @@ static PyObject *t_messageformat_mod(t_messageformat *self, PyObject *args)
         PyErr_SetObject(PyExc_TypeError, args);
         return NULL;
     }
-    
+
     STATUS_CALL(
         {
             self->object->format(f, len, _u, _fp, status);
@@ -1227,7 +1317,7 @@ static PyObject *t_pluralrules_select(t_pluralrules *self, PyObject *arg)
         u = self->object->select(d);
     else
         return PyErr_SetArgsError((PyObject *) self, "select", arg);
-        
+
     return PyUnicode_FromUnicodeString(&u);
 }
 
@@ -1717,20 +1807,190 @@ static PyObject *t_listformatter_createInstance(PyTypeObject *type,
 
 #endif
 
+#if U_ICU_VERSION_HEX >= VERSION_HEX(57, 0, 0)
+
+static int t_simpleformatter_init(t_simpleformatter *self,
+                                  PyObject *args, PyObject *kwds)
+{
+    UnicodeString *u, _u;
+    int n0, n1;
+
+    switch (PyTuple_Size(args)) {
+      case 0:
+        self->object = new SimpleFormatter();
+        self->pattern = Py_None;
+        Py_INCREF(self->pattern);
+        self->flags = T_OWNED;
+        return 0;
+
+      case 1:
+        if (!parseArgs(args, "S", &u, &_u))
+        {
+            SimpleFormatter *formatter;
+
+            INT_STATUS_CALL(formatter = new SimpleFormatter(*u, status));
+            self->object = formatter;
+            self->pattern = PyUnicode_FromUnicodeString(u);
+            self->flags = T_OWNED;
+
+            return 0;
+        }
+        break;
+
+      case 3:
+        if (!parseArgs(args, "Sii", &u, &_u, &n0, &n1))
+        {
+            SimpleFormatter *formatter;
+
+            INT_STATUS_CALL(
+                formatter = new SimpleFormatter(*u, n0, n1, status));
+            self->object = formatter;
+            self->pattern = PyUnicode_FromUnicodeString(u);
+            self->flags = T_OWNED;
+
+            return 0;
+        }
+        break;
+    }
+
+    PyErr_SetArgsError((PyObject *) self, "__init__", args);
+    return -1;
+}
+
+static PyObject *t_simpleformatter_str(t_simpleformatter *self)
+{
+    Py_INCREF(self->pattern);
+    return self->pattern;
+}
+
+static PyObject *t_simpleformatter_applyPattern(
+    t_simpleformatter *self, PyObject *arg)
+{
+    UnicodeString *u, _u;
+
+    if (!parseArg(arg, "S", &u, &_u))
+    {
+        UBool result;
+
+        STATUS_CALL(result = self->object->applyPattern(*u, status));
+        Py_DECREF(self->pattern);
+        self->pattern = PyUnicode_FromUnicodeString(u);
+
+        Py_RETURN_BOOL(result);
+    }
+
+    return PyErr_SetArgsError((PyObject *) self, "applyPattern", arg);
+}
+
+static PyObject *t_simpleformatter_applyPatternMinMaxArguments(
+    t_simpleformatter *self, PyObject *args)
+{
+    UnicodeString *u, _u;
+    int n0, n1;
+
+    switch (PyTuple_Size(args)) {
+      case 3:
+        if (!parseArgs(args, "Sii", &u, &_u, &n0, &n1))
+        {
+            UBool result;
+
+            STATUS_CALL(result = self->object->applyPatternMinMaxArguments(
+                *u, n0, n1, status));
+            Py_DECREF(self->pattern);
+            self->pattern = PyUnicode_FromUnicodeString(u);
+
+            Py_RETURN_BOOL(result);
+        }
+        break;
+    }
+
+    return PyErr_SetArgsError((PyObject *) self, "applyPatternMinMaxArguments", args);
+}
+
+static PyObject *t_simpleformatter_getArgumentLimit(
+    t_simpleformatter *self, PyObject *args)
+{
+    return PyInt_FromLong(self->object->getArgumentLimit());
+}
+
+static PyObject *t_simpleformatter_format(
+    t_simpleformatter *self, PyObject *args)
+{
+    UnicodeString *u0, *u1, *u2, _u0, _u1, _u2;
+    UnicodeString u;
+
+    switch (PyTuple_Size(args)) {
+      case 1:
+        if (!parseArgs(args, "S", &u0, &_u0))
+        {
+            STATUS_CALL(u = self->object->format(*u0, u, status));
+            return PyUnicode_FromUnicodeString(&u);
+        }
+        break;
+      case 2:
+        if (!parseArgs(args, "SS", &u0, &_u0, &u1, &_u1))
+        {
+            STATUS_CALL(u = self->object->format(*u0, *u1, u, status));
+            return PyUnicode_FromUnicodeString(&u);
+        }
+        break;
+      case 3:
+        if (!parseArgs(args, "SSS", &u0, &_u0, &u1, &_u1, &u2, &_u2))
+        {
+            STATUS_CALL(u = self->object->format(*u0, *u1, *u2, u, status));
+            return PyUnicode_FromUnicodeString(&u);
+        }
+        break;
+    }
+
+    return PyErr_SetArgsError((PyObject *) self, "format", args);
+}
+
+static PyObject *t_simpleformatter_formatStrings(
+    t_simpleformatter *self, PyObject *arg)
+{
+    UnicodeString *strings, u;
+    int count;
+
+    if (!parseArg(arg, "T", &strings, &count))
+    {
+        UnicodeString **args = new UnicodeString *[count];
+
+        for (int i = 0; i < count; ++i)
+            args[i] = &strings[i];
+
+        STATUS_CALL(
+            u = self->object->formatAndAppend(args, count, u, NULL, 0, status);
+            delete[] args;
+            delete[] strings);
+
+        return PyUnicode_FromUnicodeString(&u);
+    }
+
+    return PyErr_SetArgsError((PyObject *) self, "formatStrings", arg);
+}
+
+#endif
+
 void _init_format(PyObject *m)
 {
-    FieldPositionType.tp_richcompare = (richcmpfunc) t_fieldposition_richcmp;
-    ParsePositionType.tp_richcompare = (richcmpfunc) t_parseposition_richcmp;
-    FormatType.tp_richcompare = (richcmpfunc) t_format_richcmp;
-    MessageFormatType.tp_str = (reprfunc) t_messageformat_str;
-    MessageFormatType.tp_as_number = &t_messageformat_as_number;
-    MessageFormatType.tp_flags |= Py_TPFLAGS_CHECKTYPES;
+    FieldPositionType_.tp_richcompare = (richcmpfunc) t_fieldposition_richcmp;
+    ParsePositionType_.tp_richcompare = (richcmpfunc) t_parseposition_richcmp;
+    FormatType_.tp_richcompare = (richcmpfunc) t_format_richcmp;
+    MessageFormatType_.tp_str = (reprfunc) t_messageformat_str;
+    MessageFormatType_.tp_as_number = &t_messageformat_as_number;
+    MessageFormatType_.tp_flags |= Py_TPFLAGS_CHECKTYPES;
 #if U_ICU_VERSION_HEX >= 0x04000000
-    PluralRulesType.tp_richcompare = (richcmpfunc) t_pluralrules_richcmp;
-    PluralFormatType.tp_str = (reprfunc) t_pluralformat_str;
+    PluralRulesType_.tp_richcompare = (richcmpfunc) t_pluralrules_richcmp;
+    PluralFormatType_.tp_str = (reprfunc) t_pluralformat_str;
 #endif
 #if U_ICU_VERSION_HEX >= 0x04040000
-    SelectFormatType.tp_str = (reprfunc) t_selectformat_str;
+    SelectFormatType_.tp_str = (reprfunc) t_selectformat_str;
+#endif
+#if U_ICU_VERSION_HEX >= VERSION_HEX(57, 0, 0)
+    SimpleFormatterType_.tp_str = (reprfunc) t_simpleformatter_str;
+    SimpleFormatterType_.tp_as_number = &t_simpleformatter_as_number;
+    SimpleFormatterType_.tp_flags |= Py_TPFLAGS_CHECKTYPES;
 #endif
 
     REGISTER_TYPE(FieldPosition, m);
@@ -1751,6 +2011,9 @@ void _init_format(PyObject *m)
 #if U_ICU_VERSION_HEX >= VERSION_HEX(50, 0, 0)
     INSTALL_TYPE(ListFormatter, m);
 #endif
+#if U_ICU_VERSION_HEX >= VERSION_HEX(57, 0, 0)
+    INSTALL_STRUCT(SimpleFormatter, m);
+#endif
 
     INSTALL_STATIC_INT(FieldPosition, DONT_CARE);
 
